from BaseHTTPServer import (
    HTTPServer,
    BaseHTTPRequestHandler,
    )
import httplib
import os
from signal import SIGKILL
import simplejson
from StringIO import StringIO
from unittest import TestCase
from urlparse import urlparse
from urlparse import parse_qs

from testtools import ExpectedException

from grackle.client import (
    GrackleClient,
    )


class ForkedFake:

    def __init__(self, port, messages=None):
        self.pid = None
        self.port = port
        self.messages = messages
        self.read_end, self.write_end = os.pipe()

    def is_ready(self):
        os.write(self.write_end, 'asdf')

    def __enter__(self):
        pid = os.fork()
        if pid == 0:
            self.start_server()
        self.pid = pid
        os.read(self.read_end, 1)
        return

    def start_server(self):
        service = HTTPServer(('', self.port), FakeGrackleRequestHandler)
        service.messages = self.messages
        self.is_ready()
        service.serve_forever()

    def __exit__(self, exc_type, exc_val, traceback):
        os.kill(self.pid, SIGKILL)


class FakeGrackleRequestHandler(BaseHTTPRequestHandler):

    def do_POST(self):
        message = self.rfile.read(int(self.headers['content-length']))
        if message == 'This is a message':
            self.send_response(httplib.CREATED)
            self.end_headers()
            self.wfile.close()
        else:
            self.send_error(httplib.BAD_REQUEST)

    def do_GET(self):
        scheme, netloc, path, params, query_string, fragments = (
            urlparse(self.path))
        archive = os.path.split(path)[1]
        query = parse_qs(query_string)
        parameters = simplejson.loads(query['parameters'][0])
        self.send_response(httplib.OK)
        self.end_headers()
        messages = [m for m in self.server.messages[archive] if 'message_ids'
                    not in parameters or m['message_id'] in
                    parameters['message_ids']]
        limit = parameters.get('limit', 100)
        memo = parameters.get('memo')
        message_id_indices = dict(
            (m['message_id'], idx) for idx, m in enumerate(messages))
        if memo is None:
            start = 0
        else:
            start = message_id_indices[memo.encode('rot13')]
        if start > 0:
            previous_memo = messages[start - 1]['message_id'].encode('rot13')
        else:
            previous_memo = None
        end = min(start + limit, len(messages))
        if end < len(messages):
            next_memo = messages[end]['message_id'].encode('rot13')
        else:
            next_memo = None
        messages = messages[start:end]
        response = {
            'messages': messages,
            'next_memo': next_memo,
            'previous_memo': previous_memo
            }
        self.wfile.write(simplejson.dumps(response))


def fake_grackle_service(client, messages=None):
    if messages is None:
        messages = {}
    return ForkedFake(client.port, messages)


class TestPutMessage(TestCase):

    def test_put_message(self):
        client = GrackleClient('localhost', 8436)
        with fake_grackle_service(client):
            client.put_message('arch1', 'asdf', StringIO('This is a message'))
            with ExpectedException(Exception, 'wtf'):
                client.put_message('arch1', 'asdf',
                    StringIO('This is not a message'))


class TestGetMessages(TestCase):

    def assertMessageIDs(self, ids, messages):
        self.assertEqual(
            sorted(ids), sorted(m['message_id'] for m in messages))

    def test_get_messages(self):
        client = GrackleClient('localhost', 8435)
        with fake_grackle_service(client,
            {'baz':
            [{'message_id': 'foo'},
             {'message_id': 'bar'}]}):
            response = client.get_messages('baz')
        self.assertEqual(['bar', 'foo'], sorted(m['message_id'] for m in
            response['messages']))
        self.assertIs(None, response['next_memo'])
        self.assertIs(None, response['previous_memo'])

    def test_get_messages_by_id(self):
        client = GrackleClient('localhost', 8437)
        with fake_grackle_service(client,
            {'baz':
            [{'message_id': 'foo'},
             {'message_id': 'bar'}]}):
            response = client.get_messages('baz', message_ids=['foo'])
        message, = response['messages']
        self.assertEqual('foo', message['message_id'])

    def test_get_messages_batching(self):
        client = GrackleClient('localhost', 8437)
        with fake_grackle_service(client,
            {'baz':
            [{'message_id': 'foo'},
             {'message_id': 'bar'}]}):
            response = client.get_messages('baz', limit=1)
            self.assertEqual(1, len(response['messages']))
            messages = response['messages']
            response = client.get_messages(
                'baz', limit=1, memo=response['next_memo'])
            self.assertEqual(1, len(response['messages']))
            messages.extend(response['messages'])
            self.assertMessageIDs(['foo', 'bar'], messages)
